"""
Utilities for doing light scrapping of Rotten Tomatoes
to get additional movie reviews with their ratings.
"""
def get_box_office_links():
    import requests
    import lxml.html
    url = "https://www.rottentomatoes.com"
    page = requests.get(url)
    doc = lxml.html.document_fromstring(page.content)
    n = doc.cssselect("#homepage-top-box-office .movie_list")
    links = set()
    for child in n[0].cssselect("a"):
        links.add(url + child.attrib["href"])
    return links


def more_reviews_available(document, pagenumber):
    button_links = document.cssselect(".btn.btn-xs.btn-primary-rt")
    found = False
    desirata = "/reviews/?page=%d&type=user" % (pagenumber,)
    for blink in button_links:
        if desirata in blink.attrib["href"]:
            found = True
    return found


def get_reviews_for_url(url, page_num=0):
    import requests
    import lxml.html
    page = requests.get(url + "/reviews/?page=%d&type=user" % (page_num,))
    doc = lxml.html.document_fromstring(page.content)
    reviews_nodes = doc.cssselect(".user_review")
    reviews = []
    for rev in reviews_nodes:
        score_node = rev.cssselect(".scoreWrapper span")[0]
        try:
            score = float(score_node.attrib["class"]) / 50.0
        except ValueError:
            # review is a want to see it, skip
            continue
        try:
            text = "\n".join(list(rev.itertext())).strip()
            reviews.append((score, text))
        except UnicodeDecodeError:
            continue
    if more_reviews_available(doc, page_num+1):
        reviews += get_reviews_for_url(url, page_num+1)
    return reviews


def get_box_office_reviews():
    links = get_box_office_links()
    all_reviews = []
    for link in links:
        all_reviews.extend(get_reviews_for_url(link))
    return all_reviews




"""
Make trees visualizable in an IPython notebook
"""
import json

try:
    from PIL import ImageFont
    font = ImageFont.core.getfont("/Library/Fonts/Georgia.ttf", 15)
    def text_size(text):
        return max(4, font.getsize(text)[0][0])
except Exception:
    def text_size(text):
        # TODO(contributors): make changes here to incorporate cap and uncap unknown words.
        return max(4, int(len(text) * 1.1))


class LabeledSentence(object):
    __slots__ = ["text", "label"]
    def __init__(self, text):
        self.text = text
        self.label    = None


class LabeledTree(object):
    SCORE_MAPPING = [-12.5,-6.25,0.0,6.25,12.5]

    def __init__(self,
                 depth=0,
                 text=None,
                 label=None,
                 children=None,
                 parent=None,
                 udepth=1):
        self.label    = label
        self.children = children if children != None else []
        self.general_children = []
        self.text = text
        self.parent   = parent
        self.depth    = depth
        self.udepth   = udepth

    def uproot(tree):
        """
        Take a subranch of a tree and deep-copy the children
        of this subbranch into a new LabeledTree
        """
        uprooted = tree.copy()
        uprooted.parent = None
        for child in tree.all_children():
            uprooted.add_general_child(child)
        return uprooted

    def shrink_tree(tree, final_depth):
        if tree.udepth <= final_depth:
            return tree
        for branch in tree.general_children:
            if branch.udepth == final_depth:
                return branch.uproot()

    def shrunk_trees(tree, final_depth):
        if tree.udepth <= final_depth:
            yield tree
        for branch in tree.general_children:
            if branch.udepth == final_depth:
                yield branch.uproot()

    def copy(self):
        """
        Deep Copy of a LabeledTree
        """
        return LabeledTree(
            udepth = self.udepth,
            depth = self.depth,
            text = self.text,
            label = self.label,
            children = self.children.copy() if self.children != None else [],
            parent = self.parent)

    def add_child(self, child):
        """
        Adds a branch to the current tree.
        """
        self.children.append(child)
        child.parent = self
        self.udepth = max([child.udepth for child in self.children]) + 1

    def add_general_child(self, child):
        self.general_children.append(child)

    def all_children(self):
        if len(self.children) > 0:
            for child in self.children:
                for subchild in child.all_children():
                    yield subchild
            yield self
        else:
            yield self

    def lowercase(self):
        """
        Lowercase all strings in this tree.
        Works recursively and in-place.
        """
        if len(self.children) > 0:
            for child in self.children:
                child.lowercase()
        else:
            self.text = self.text.lower()

    def to_dict(self, index=0):
        """
        Dict format for use in Javascript / Jason Chuang's display technology.
        """
        index += 1
        rep = {}
        rep["index"] = index
        rep["leaf"] = len(self.children) == 0
        rep["depth"] = self.udepth
        rep["scoreDistr"] = [0.0] * len(LabeledTree.SCORE_MAPPING)
        # dirac distribution at correct label
        if self.label is not None:
            rep["scoreDistr"][self.label] = 1.0
            mapping = LabeledTree.SCORE_MAPPING[:]
            rep["rating"] = mapping[self.label] - min(mapping)
        # if you are using this method for printing predictions
        # from a model, the the dot product with the model's output
        # distribution should be taken with this list:
        rep["numChildren"] = len(self.children)
        text = self.text if self.text != None else ""
        seen_tokens = 0
        witnessed_pixels = 0
        for i, child in enumerate(self.children):
            if i > 0:
                text += " "
            child_key = "child%d" % (i)
            (rep[child_key], index) = child.to_dict(index)
            text += rep[child_key]["text"]
            seen_tokens += rep[child_key]["tokens"]
            witnessed_pixels += rep[child_key]["pixels"]

        rep["text"] = text
        rep["tokens"] = 1 if (self.text != None and len(self.text) > 0) else seen_tokens
        rep["pixels"] = witnessed_pixels + 3 if len(self.children) > 0 else text_size(self.text)
        return (rep, index)

    def to_json(self):
        rep, _ = self.to_dict()
        return json.dumps(rep)

    def display(self):
        from IPython.display import Javascript, display

        display(Javascript("createTrees(["+self.to_json()+"])"))
        display(Javascript("updateTrees()"))

    def to_lines(self):
        if len(self.children) > 0:
            left_lines, right_lines = self.children[0].to_lines(), self.children[1].to_lines()
            self_line = [left_lines[0] + " " + right_lines[0]]
            return self_line + left_lines + right_lines
        else:
            return [self.text]

    def to_labeled_lines(self):
        if len(self.children) > 0:
            left_lines, right_lines = self.children[0].to_labeled_lines(), self.children[1].to_labeled_lines()
            self_line = [(self.label, left_lines[0][1] + " " + right_lines[0][1])]
            return self_line + left_lines + right_lines
        else:
            return [(self.label, self.text)]

    def __str__(self):
        """
        String representation of a tree as visible in original corpus.

        print(tree)
        #=> '(2 (2 not) (3 good))'

        Outputs
        -------

            str: the String representation of the tree.

        """
        if len(self.children) > 0:
            rep = "(%d " % self.label
            for child in self.children:
                rep += str(child)
            return rep + ")"
        else:
            text = self.text\
                .replace("(", "-LRB-")\
                .replace(")", "-RRB-")\
                .replace("{", "-LCB-")\
                .replace("}", "-RCB-")\
                .replace("[", "-LSB-")\
                .replace("]", "-RSB-")

            return ("(%d %s) " % (self.label, text))

    @staticmethod
    def inject_visualization_javascript(tree_width=1200, tree_height=400, tree_node_radius=10):
        """
        In an Ipython notebook, show SST trees using the same Javascript
        code as used by Jason Chuang's visualisations.
        """
        from .javascript import insert_sentiment_markup
        insert_sentiment_markup(tree_width=tree_width, tree_height=tree_height, tree_node_radius=tree_node_radius)



from __future__ import print_function

import random
import sys
import subprocess

from os import remove, rmdir, stat
from os.path import join, dirname, realpath, exists, isdir, isfile

from .utils import print

def execute_bash(command):
    """
    Executes bash command, prints output and throws an exception on failure.

    Arguments:
    ----------
        command : str, bash command to run.

    """
    process = subprocess.Popen(command,
                               shell=True,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT,
                               universal_newlines=True)
    for line in process.stdout:
        print(line, end='', flush=True)
    process.wait()
    assert process.returncode == 0

def delete_paths(paths):
    """
    Delete a list of paths that are files or directories.
    If a file/directory does not exist, skip it.

    Arguments:
    ----------

    paths : list<str>, names of files/directories to remove.

    """
    for path in paths:
        if exists(path):
            if isfile(path):
                remove(path)
            elif isdir(path):
                rmdir(path)

def download_sst(path, url):
    """"
    Download from `url` the zip file corresponding to the
    Stanford Sentiment Treebank and expand the resulting
    files into the directory `path` (Note: if the files are
    already present, the download is not actually run).

    Arguments
    ---------
        path : str, directory to save the train, test, and dev files.
        url : str, location of zip file on the web

    Returns:
    --------

        dict<str, str>: file path for the keys train, test, dev.

    """
    local_files = {
        "train": join(path, "train.txt"),
        "test": join(path, "test.txt"),
        "dev": join(path, "dev.txt")
    }
    if all(exists(fname) and stat(fname).st_size > 100 for fname in local_files.values()):
        return local_files

    zip_local = join(path, 'trainDevTestTrees_PTB.zip')
    delete_paths([zip_local, join(path, "trees")] + list(local_files.values()))
    execute_bash('wget -O %s %s' % (zip_local, url))
    execute_bash('unzip %s -d %s' % (zip_local, path))
    execute_bash('mv %s %s' % (join(path, "trees", "*"), path))
    delete_paths([zip_local, join(path, "trees")])
    return local_files



"""
Special loading methods for importing dataset as processed
by the TreeLSTM code from https://github.com/stanfordnlp/treelstm
"""
from .labeled_trees import LabeledTree
import codecs

def import_tree_corpus(labels_path, parents_path, texts_path):
    """
    Import dataset from the TreeLSTM data generation scrips.

    Inputs
    ------

    labels_path: where are labels are stored (should be in data/sst/labels.txt)
    parents_path: where the parent relationships are stored (should be in data/sst/parents.txt)
    texts_path: where are strings for each tree are stored (should be in data/sst/sents.txt)

    Outputs
    -------

    trees : a list of LabeledTree
    """
    with codecs.open(labels_path, "r", "UTF-8") as f:
        label_lines = f.readlines()
    with codecs.open(parents_path, "r", "UTF-8") as f:
        parent_lines = f.readlines()
    with codecs.open(texts_path, "r", "UTF-8") as f:
        word_lines = f.readlines()
    assert len(label_lines) == len(parent_lines)
    assert len(label_lines) == len(word_lines)

    trees = []

    for labels, parents, words in zip(label_lines, parent_lines, word_lines):
        labels  = [int(l) + 2 for l in labels.strip().split(" ")]
        parents = [int(l) for l in parents.strip().split(" ")]
        words   = words.strip().split(" ")
        assert len(labels) == len(parents)
        trees.append(read_tree(parents, labels, words))
    return trees

def assign_texts(node, words, next_idx=0):
    """
    Recursively assign the words to nodes by finding and
    assigning strings to the leaves of a tree in left
    to right order.
    """
    if len(node.children) == 0:
        node.text = words[next_idx]
        return next_idx + 1
    else:
        for child in node.children:
            next_idx = assign_texts(child, words, next_idx)
        return next_idx

def read_tree(parents, labels, words):
    """
    Take as input a list of integers for parents
    and labels, along with a list of words, and
    reconstruct a LabeledTree.
    """
    trees = {}
    root = None
    for i in range(1, len(parents) + 1):
        if not i in trees and parents[i - 1] != - 1:
            idx = i
            prev = None
            while True:
                parent = parents[idx - 1]
                if parent == -1:
                    break
                tree = LabeledTree()
                if prev is not None:
                    tree.add_child(prev)
                trees[idx] = tree
                tree.label = labels[idx - 1]
                if trees.get(parent) is not None:
                    trees[parent].add_child(tree)
                    break
                elif parent == 0:
                    root = tree
                    break
                else:
                    prev = tree
                    idx = parent
    assert assign_texts(root, words) == len(words)
    return root



"""
Python 3 / 2 compatibility functions.
"""
from __future__ import print_function

import os
import sys

if sys.version_info >= (3,3):
    makedirs = os.makedirs
    def normalize_string(string):
        return string.replace("\xa0", " ")\
                     .replace("\\", "")\
                     .replace("-LRB-", "(")\
                     .replace("-RRB-", ")")\
                     .replace("-LCB-", "{")\
                     .replace("-RCB-", "}")\
                     .replace("-LSB-", "[")\
                     .replace("-RSB-", "]")
    old_print = print
    def print(*args, **kwargs):
        old_print(*args, **kwargs)
else:
    old_makedirs = os.makedirs
    def makedirs(path, mode=0o777, exist_ok=False):
        if not exist_ok:
            return old_makedirs(path, mode)
        else:
            if not os.path.exists(path):
                return old_makedirs(path, mode)

    def normalize_string(string):
        return string.replace(u"\xa0", " ")\
                     .replace("\\", "")\
                     .replace("-LRB-", "(")\
                     .replace("-RRB-", ")")\
                     .replace("-LCB-", "{")\
                     .replace("-RCB-", "}")\
                     .replace("-LSB-", "[")\
                     .replace("-RSB-", "]")

    old_print = print
    def print(*args, **kwargs):
        flush = kwargs.pop('flush', False)
        old_print(*args, **kwargs)
        file = kwargs.get('file', sys.stdout)
        if flush and file is not None:
            file.flush()

__all__ = ["makedirs", "normalize_string"]



import os
from setuptools import setup, find_packages

def readfile(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='pytreebank',
    version='0.2.3',
    description='Python package for loading Stanford Sentiment Treebank corpus',
    long_description=readfile('README.md'),
    ext_modules=[],
    packages=find_packages(),
    author='Jonathan Raiman',
    author_email='jonathanraiman@gmail.com',
    url='https://github.com/JonathanRaiman/pytreebank',
    download_url='https://github.com/JonathanRaiman/pytreebank',
    keywords='Machine Learning, NLP',
    license='MIT',
    platforms='any',
    zip_safe=False,
    classifiers=[
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 2.7',
        'Topic :: Text Processing :: Linguistic',
    ],
    setup_requires = [],
    install_requires=[],
    include_package_data=True,
)



import codecs
import os

from collections import OrderedDict

from .labeled_trees import LabeledTree
from .download import download_sst
from .utils import makedirs, normalize_string

class ParseError(ValueError):
    pass

def attribute_text_label(node, current_word):
    node.text = normalize_string(current_word)
    node.text = node.text.strip(" ")
    node.udepth = 1
    if len(node.text) > 0 and node.text[0].isdigit():
        split_sent = node.text.split(" ", 1)
        label = split_sent[0]
        if len(split_sent) > 1:
            text = split_sent[1]
            node.text = text

        if all(c.isdigit() for c in label):
            node.label = int(label)
        else:
            text = label + " " + text
            node.text = text

    if len(node.text) == 0:
        node.text = None


def create_tree_from_string(line):
    depth         = 0
    current_word  = ""
    root          = None
    current_node  = root

    for char in line:
        if char == '(':
            if current_node is not None and len(current_word) > 0:
                attribute_text_label(current_node, current_word)
                current_word = ""
            depth += 1
            if depth > 1:
                # replace current head node by this node:
                child = LabeledTree(depth=depth)
                current_node.add_child(child)
                current_node = child
                root.add_general_child(child)
            else:
                root = LabeledTree(depth=depth)
                root.add_general_child(root)
                current_node = root

        elif char == ')':
            # assign current word:
            if len(current_word) > 0:
                attribute_text_label(current_node, current_word)
                current_word = ""

            # go up a level:
            depth -= 1
            if current_node.parent != None:
                current_node.parent.udepth = max(current_node.udepth+1, current_node.parent.udepth)
            current_node = current_node.parent
        else:
            # add to current read word
            current_word += char
    if depth != 0:
        raise ParseError("Not an equal amount of closing and opening parentheses")

    return root



class LabeledTreeCorpus(list):
    """
    Read in the Stanford Sentiment Treebank using the original serialization format:

    > (3 (2 this) (3 (2 is) (3 good ) )

    """
    def labels(self):
        labelings = OrderedDict()
        for tree in self:
            for label, line in tree.to_labeled_lines():
                labelings[line] = label
        return labelings

    def to_file(self, path, mode = "w"):
        with open(path, mode = mode) as f:
            for tree in self:
                for label, line in tree.to_labeled_lines():
                    f.write(line + "\n")



def import_tree_corpus(trees):
    tree_list = LabeledTreeCorpus()
    with codecs.open(trees, "r", "UTF-8") as f:
        for line in f:
            tree_list.append(create_tree_from_string(line))
    return tree_list


def load_sst(path=None,
             url='http://nlp.stanford.edu/sentiment/trainDevTestTrees_PTB.zip'):
    if path is None:
        # find a good temporary path
        path = os.path.expanduser("~/stanford_sentiment_treebank/")
        makedirs(path, exist_ok=True)
    fnames = download_sst(path, url)
    return {key: import_tree_corpus(value) for key, value in fnames.items()}




from os.path import realpath, dirname, join
from IPython.display import HTML, display

SCRIPT_DIR = dirname(realpath(__file__))

def import_tag(tagname, contents = "", **kwargs):
	html = "<%s " % (tagname)
	for key, val in kwargs.items():
		if key == 'className':
			key = 'class' # to avoid Python issues
		html += "%s='%s' " % (key, val)
	html += ">"
	html += contents
	html += "</%s>" % (tagname)
	display(HTML(html))

def format_replacements(content, replacements):
	for key, value in replacements.items():
		content = content.replace("{" + key + "}", str(value))
	return content

def import_javascript(scriptname, replacements=None):
	if replacements is None:
		replacements = {}
	import_tag(
		"script",
		contents=format_replacements(open(scriptname).read(), replacements),
		type="text/javascript"
	)

def import_css(cssname):
	import_tag("style", contents=open(cssname).read())

def insert_stanford_javascript(tree_width=1200, tree_height=400, tree_node_radius=10):
	import_javascript(join(SCRIPT_DIR, "c3.min.js"))
	import_javascript(join(SCRIPT_DIR, "d3.min.js"))
	import_javascript(join(SCRIPT_DIR, "sentree.min.js"))
	import_javascript(join(SCRIPT_DIR, "sentrees.min.js"),
		{
			"treeWidth":tree_width,
			"treeHeight":tree_height,
			"treeNodeRadius":tree_node_radius
		})
	import_javascript(join(SCRIPT_DIR, "tree_visualization.js"))

def insert_stanford_styles():
	import_css(join(SCRIPT_DIR, "tree_visualization.css"))

def insert_sentiment_markup(tree_width=1200, tree_height=400, tree_node_radius=10):
	insert_stanford_javascript(
		tree_width=tree_width,
		tree_height=tree_height,
		tree_node_radius=tree_node_radius
	)
	insert_stanford_styles()
	import_tag("div", className='trees')
